"""Device Tracker platform for zhawss."""
from __future__ import annotations

import asyncio
import functools
import time
from typing import TYPE_CHECKING

from zhaws.server.decorators import periodic
from zhaws.server.platforms import PlatformEntity
from zhaws.server.platforms.registries import PLATFORM_ENTITIES, Platform
from zhaws.server.platforms.sensor import Battery
from zhaws.server.zigbee.cluster import (
    CLUSTER_HANDLER_EVENT,
    ClusterAttributeUpdatedEvent,
)
from zhaws.server.zigbee.cluster.const import CLUSTER_HANDLER_POWER_CONFIGURATION

if TYPE_CHECKING:
    from zhaws.server.zigbee.cluster import ClusterHandler
    from zhaws.server.zigbee.device import Device
    from zhaws.server.zigbee.endpoint import Endpoint

STRICT_MATCH = functools.partial(
    PLATFORM_ENTITIES.strict_match, Platform.DEVICE_TRACKER
)


@STRICT_MATCH(cluster_handler_names=CLUSTER_HANDLER_POWER_CONFIGURATION)
class DeviceTracker(PlatformEntity):
    """Representation of a zhawss device tracker."""

    PLATFORM = Platform.DEVICE_TRACKER

    def __init__(
        self,
        unique_id: str,
        cluster_handlers: list[ClusterHandler],
        endpoint: Endpoint,
        device: Device,
    ):
        """Initialize the binary sensor."""
        super().__init__(unique_id, cluster_handlers, endpoint, device)
        self._battery_cluster_handler: ClusterHandler = self.cluster_handlers[
            CLUSTER_HANDLER_POWER_CONFIGURATION
        ]
        self._connected: bool = False
        self._keepalive_interval: int = 60
        self._should_poll: bool = True
        self._battery_level: float | None = None
        self._battery_cluster_handler.on_event(
            CLUSTER_HANDLER_EVENT, self._handle_event_protocol
        )
        self._cancel_refresh_handle = asyncio.create_task(self._refresh())

    async def async_update(self) -> None:
        """Handle polling."""
        if self.device.last_seen is None:
            self._connected = False
        else:
            difference = time.time() - self.device.last_seen
            if difference > self._keepalive_interval:
                self._connected = False
            else:
                self._connected = True
        self.maybe_send_state_changed_event()

    @periodic((30, 45))
    async def _refresh(self) -> None:
        """Refresh the state of the device tracker."""
        await self.async_update()

    @property
    def is_connected(self) -> bool:
        """Return true if the device is connected to the network."""
        return self._connected

    def handle_cluster_handler_attribute_updated(
        self, event: ClusterAttributeUpdatedEvent
    ) -> None:
        """Handle tracking."""
        if event.name != "battery_percentage_remaining":
            return
        self.debug("battery_percentage_remaining updated: %s", event.value)
        self._connected = True
        self._battery_level = Battery.formatter(event.value)
        self.maybe_send_state_changed_event()

    @property
    def battery_level(self) -> float | None:
        """Return the battery level of the device.
        Percentage from 0-100.
        """
        return self._battery_level

    def get_state(self) -> dict:
        """Return the state of the device."""
        response = super().get_state()
        response.update(
            {
                "connected": self._connected,
                "battery_level": self._battery_level,
            }
        )
        return response



"""Device discovery functions for Zigbee Home Automation."""
from __future__ import annotations

from collections import Counter
import logging
from typing import TYPE_CHECKING, Any

from zhaws.server.platforms import (  # noqa: F401 pylint: disable=unused-import,
    GroupEntity,
    alarm_control_panel,
    binary_sensor,
    button,
    climate,
    cover,
    device_tracker,
    fan,
    light,
    lock,
    number,
    select,
    sensor,
    siren,
    switch,
)
from zhaws.server.platforms.registries import (
    DEVICE_CLASS,
    PLATFORM_ENTITIES,
    REMOTE_DEVICE_TYPES,
    SINGLE_INPUT_CLUSTER_DEVICE_CLASS,
    SINGLE_OUTPUT_CLUSTER_DEVICE_CLASS,
    Platform,
)

if TYPE_CHECKING:
    from zhaws.server.websocket.server import Server
    from zhaws.server.zigbee.endpoint import Endpoint
    from zhaws.server.zigbee.group import Group
    from zhaws.server.zigbee.controller import Controller

from zhaws.server.zigbee.cluster import (  # noqa: F401
    ClusterHandler,
    closures,
    general,
    homeautomation,
    hvac,
    lighting,
    lightlink,
    manufacturerspecific,
    measurement,
    protocol,
    security,
    smartenergy,
)
from zhaws.server.zigbee.registries import (
    CLUSTER_HANDLER_REGISTRY,
    HANDLER_ONLY_CLUSTERS,
)

_LOGGER = logging.getLogger(__name__)


PLATFORMS = (
    Platform.ALARM_CONTROL_PANEL,
    Platform.BINARY_SENSOR,
    Platform.BUTTON,
    Platform.CLIMATE,
    Platform.COVER,
    Platform.DEVICE_TRACKER,
    Platform.FAN,
    Platform.LIGHT,
    Platform.LOCK,
    Platform.NUMBER,
    Platform.SELECT,
    Platform.SENSOR,
    Platform.SIREN,
    Platform.SWITCH,
)

GROUP_PLATFORMS = (
    Platform.FAN,
    Platform.LIGHT,
    Platform.SWITCH,
)


class ProbeEndpoint:
    """All discovered cluster handlers and entities of an endpoint."""

    def __init__(self) -> None:
        """Initialize instance."""
        self._device_configs: dict[str, Any] = {}

    def discover_entities(self, endpoint: Endpoint) -> None:
        """Process an endpoint on a zigpy device."""
        self.discover_by_device_type(endpoint)
        self.discover_multi_entities(endpoint)
        self.discover_by_cluster_id(endpoint)
        PLATFORM_ENTITIES.clean_up()

    def discover_by_device_type(self, endpoint: Endpoint) -> None:
        """Process an endpoint on a zigpy device."""

        unique_id = endpoint.unique_id

        platform = None  # remove this when the below is uncommented
        """TODO
        platform = self._device_configs.get(unique_id, {}).get(ha_const.CONF_TYPE)
        """
        if platform is None:
            ep_profile_id = endpoint.zigpy_endpoint.profile_id
            ep_device_type = endpoint.zigpy_endpoint.device_type
            platform = DEVICE_CLASS[ep_profile_id].get(ep_device_type)

        if platform and platform in PLATFORMS:
            cluster_handlers = endpoint.unclaimed_cluster_handlers()
            platform_entity_class, claimed = PLATFORM_ENTITIES.get_entity(
                platform,
                endpoint.device.manufacturer,
                endpoint.device.model,
                cluster_handlers,
            )
            if platform_entity_class is None:
                return
            endpoint.claim_cluster_handlers(claimed)
            endpoint.async_new_entity(
                platform, platform_entity_class, unique_id, claimed
            )

    def discover_by_cluster_id(self, endpoint: Endpoint) -> None:
        """Process an endpoint on a zigpy device."""

        items = SINGLE_INPUT_CLUSTER_DEVICE_CLASS.items()
        single_input_clusters = {
            cluster_class: match
            for cluster_class, match in items
            if not isinstance(cluster_class, int)
        }
        remaining_cluster_handlers = endpoint.unclaimed_cluster_handlers()
        for cluster_handler in remaining_cluster_handlers:
            if cluster_handler.cluster.cluster_id in HANDLER_ONLY_CLUSTERS:
                endpoint.claim_cluster_handlers([cluster_handler])
                continue

            platform = SINGLE_INPUT_CLUSTER_DEVICE_CLASS.get(
                cluster_handler.cluster.cluster_id
            )
            if platform is None:
                for cluster_class, match in single_input_clusters.items():
                    if isinstance(cluster_handler.cluster, cluster_class):
                        platform = match
                        break
            if platform is not None:
                self.probe_single_cluster(platform, cluster_handler, endpoint)

        # until we can get rid off registries
        self.handle_on_off_output_cluster_exception(endpoint)

    @staticmethod
    def probe_single_cluster(
        platform: Platform,
        cluster_handler: ClusterHandler,
        endpoint: Endpoint,
    ) -> None:
        """Probe specified cluster for specific platform."""
        if platform is None or platform not in PLATFORMS:
            return
        cluster_handler_list = [cluster_handler]
        unique_id = f"{endpoint.unique_id}-{cluster_handler.cluster.cluster_id}"

        entity_class, claimed = PLATFORM_ENTITIES.get_entity(
            platform,
            endpoint.device.manufacturer,
            endpoint.device.model,
            cluster_handler_list,
        )
        if entity_class is None:
            return
        endpoint.claim_cluster_handlers(claimed)
        endpoint.async_new_entity(platform, entity_class, unique_id, claimed)

    def handle_on_off_output_cluster_exception(
        self,
        endpoint: Endpoint,
    ) -> None:
        """Process output clusters of the endpoint."""

        profile_id = endpoint.zigpy_endpoint.profile_id
        device_type = endpoint.zigpy_endpoint.device_type
        if device_type in REMOTE_DEVICE_TYPES.get(profile_id, []):
            return

        for cluster_id, cluster in endpoint.zigpy_endpoint.out_clusters.items():
            platform = SINGLE_OUTPUT_CLUSTER_DEVICE_CLASS.get(cluster.cluster_id)
            if platform is None:
                continue

            cluster_handler_class = CLUSTER_HANDLER_REGISTRY.get(
                cluster_id, ClusterHandler
            )
            cluster_handler = cluster_handler_class(cluster, endpoint)
            self.probe_single_cluster(platform, cluster_handler, endpoint)

    @staticmethod
    def discover_multi_entities(endpoint: Endpoint) -> None:
        """Process an endpoint on and discover multiple entities."""

        ep_profile_id = endpoint.zigpy_endpoint.profile_id
        ep_device_type = endpoint.zigpy_endpoint.device_type
        platform_by_dev_type = DEVICE_CLASS[ep_profile_id].get(ep_device_type)
        remaining_cluster_handlers = endpoint.unclaimed_cluster_handlers()

        matches, claimed = PLATFORM_ENTITIES.get_multi_entity(
            endpoint.device.manufacturer,
            endpoint.device.model,
            remaining_cluster_handlers,
        )

        endpoint.claim_cluster_handlers(claimed)
        for platform, ent_n_handler_list in matches.items():
            for entity_and_handler in ent_n_handler_list:
                _LOGGER.debug(
                    "'%s' platform -> '%s' using %s",
                    platform,
                    entity_and_handler.entity_class.__name__,
                    [ch.name for ch in entity_and_handler.claimed_cluster_handlers],
                )
        for platform, ent_n_handler_list in matches.items():
            for entity_and_handler in ent_n_handler_list:
                if platform == platform_by_dev_type:
                    # for well known device types, like thermostats we'll take only 1st class
                    endpoint.async_new_entity(
                        platform,
                        entity_and_handler.entity_class,
                        endpoint.unique_id,
                        entity_and_handler.claimed_cluster_handlers,
                    )
                    break
                first_ch = entity_and_handler.claimed_cluster_handlers[0]
                endpoint.async_new_entity(
                    platform,
                    entity_and_handler.entity_class,
                    f"{endpoint.unique_id}-{first_ch.cluster.cluster_id}",
                    entity_and_handler.claimed_cluster_handlers,
                )

    def initialize(self, server: Server) -> None:
        """Update device overrides config."""
        """TODO
        zha_config = server.data[zha_const.DATA_ZHA].get(zha_const.DATA_ZHA_CONFIG, {})
        if overrides := zha_config.get(zha_const.CONF_DEVICE_CONFIG):
            self._device_configs.update(overrides)
        """


class GroupProbe:
    """Determine the appropriate platform for a group"""

    def __init__(self) -> None:
        """Initialize instance."""
        self._server: Server | None = None

    def initialize(self, server: Server) -> None:
        """Initialize the group probe."""
        self._server = server

    def _reprobe_group(self, group_id: int) -> None:
        """Reprobe a group for entities after its members change."""
        assert self._server is not None
        controller: Controller = self._server.controller
        if (group := controller.groups.get(group_id)) is None:
            return
        self.discover_group_entities(group)

    def discover_group_entities(self, group: Group) -> None:
        """Process a group and create any entities that are needed."""
        _LOGGER.info("Probing group %s for entities", group.name)
        # only create a group entity if there are 2 or more members in a group
        if len(group.members) < 2:
            _LOGGER.debug(
                "Group: %s:0x%04x has less than 2 members - skipping entity discovery",
                group.name,
                group.group_id,
            )
            return

        assert self._server is not None
        entity_platforms = GroupProbe.determine_entity_platforms(self._server, group)

        if not entity_platforms:
            _LOGGER.info("No entity platforms discovered for group %s", group.name)
            return

        for platform in entity_platforms:
            entity_class = PLATFORM_ENTITIES.get_group_entity(platform)
            if entity_class is None:
                continue
            _LOGGER.info("Creating entity : %s for group %s", entity_class, group.name)
            entity_class(group)

    @staticmethod
    def determine_entity_platforms(server: Server, group: Group) -> list[Platform]:
        """Determine the entity platforms for this group."""
        entity_domains: list[Platform] = []
        all_platform_occurrences = []
        for member in group.members:
            if member.device.is_coordinator:
                continue
            entities = member.associated_entities
            all_platform_occurrences.extend(
                [
                    entity.PLATFORM
                    for entity in entities
                    if entity.PLATFORM in GROUP_PLATFORMS
                ]
            )
        if not all_platform_occurrences:
            return entity_domains
        # get all platforms we care about if there are more than 2 entities of this platform
        counts = Counter(all_platform_occurrences)
        entity_platforms = [
            platform[0] for platform in counts.items() if platform[1] >= 2
        ]
        _LOGGER.debug(
            "The entity platforms are: %s for group: %s:0x%04x",
            entity_platforms,
            group.name,
            group.group_id,
        )
        return entity_platforms


PROBE: ProbeEndpoint = ProbeEndpoint()
GROUP_PROBE: GroupProbe = GroupProbe()



"""Device for the zhaws.client."""
from __future__ import annotations

from typing import TYPE_CHECKING

from zhaws.event import EventBase

if TYPE_CHECKING:
    from zhaws.client.client import Client
    from zhaws.client.controller import Controller
    from zhaws.client.model.types import Device as DeviceModel


class Device(EventBase):
    """Device for the zhaws.client."""

    def __init__(self, device: DeviceModel, controller: Controller, client: Client):
        """Initialize the Device class."""
        super().__init__()
        self._device: DeviceModel = device
        self._controller: Controller = controller
        self._client: Client = client

    @property
    def device(self) -> DeviceModel:
        """Return the device."""
        return self._device

    @device.setter
    def device(self, device: DeviceModel) -> None:
        """Set the device."""
        self._device = device

    @property
    def controller(self) -> Controller:
        """Return the controller."""
        return self._controller

    @property
    def client(self) -> Client:
        """Return the client."""
        return self._client

    def __repr__(self) -> str:
        return self._device.__repr__()



"""Utils for zhawss."""
from __future__ import annotations

from typing import TYPE_CHECKING, Any

from zigpy.zcl import Cluster as ZigpyCluster

if TYPE_CHECKING:
    from zhaws.server.zigbee.cluster import ClusterHandler


async def safe_read(
    cluster: ZigpyCluster,
    attributes: list,
    allow_cache: bool = True,
    only_cache: bool = False,
    manufacturer: int | None = None,
) -> dict:
    """Swallow all exceptions from network read.
    If we throw during initialization, setup fails. Rather have an entity that
    exists, but is in a maybe wrong state, than no entity. This method should
    probably only be used during initialization.
    """
    try:
        result, _ = await cluster.read_attributes(
            attributes,
            allow_cache=allow_cache,
            only_cache=only_cache,
            manufacturer=manufacturer,
        )
        return result
    except Exception:  # pylint: disable=broad-except
        return {}


def parse_and_log_command(
    cluster_handler: ClusterHandler, tsn: int, command_id: int, args: Any
) -> str:
    """Parse and log a zigbee cluster command."""
    cmd: str = cluster_handler.cluster.server_commands.get(command_id, [command_id])[0]
    cluster_handler.debug(
        "received '%s' command with %s args on cluster_id '%s' tsn '%s'",
        cmd,
        args,
        cluster_handler.cluster.cluster_id,
        tsn,
    )
    return cmd



"""Group for the zhaws.client."""
from __future__ import annotations

from typing import TYPE_CHECKING

from zhaws.event import EventBase

if TYPE_CHECKING:
    from zhaws.client.client import Client
    from zhaws.client.controller import Controller
    from zhaws.client.model.types import Group as GroupModel


class Group(EventBase):
    """Group for the zhaws.client."""

    def __init__(self, group: GroupModel, controller: Controller, client: Client):
        """Initialize the Group class."""
        super().__init__()
        self._group: GroupModel = group
        self._controller: Controller = controller
        self._client: Client = client

    @property
    def group(self) -> GroupModel:
        """Return the group."""
        return self._group

    @group.setter
    def group(self, group: GroupModel) -> None:
        """Set the group."""
        self._group = group

    @property
    def controller(self) -> Controller:
        """Return the controller."""
        return self._controller

    @property
    def client(self) -> Client:
        """Return the client."""
        return self._client

    def __repr__(self) -> str:
        return self._group.__repr__()



"""Light platform for zhawss."""
from __future__ import annotations

import asyncio
from collections import Counter
import enum
import functools
import itertools
from typing import TYPE_CHECKING, Any, Final, Optional, Union

from zigpy.zcl.clusters.general import Identify, LevelControl, OnOff
from zigpy.zcl.clusters.lighting import Color
from zigpy.zcl.foundation import Status

from zhaws.server.decorators import periodic
from zhaws.server.platforms import BaseEntity, GroupEntity, PlatformEntity, helpers
from zhaws.server.platforms.registries import PLATFORM_ENTITIES, Platform
from zhaws.server.platforms.util import color as color_util
from zhaws.server.zigbee.cluster import (
    CLUSTER_HANDLER_EVENT,
    ClusterAttributeUpdatedEvent,
)
from zhaws.server.zigbee.cluster.const import (
    CLUSTER_HANDLER_COLOR,
    CLUSTER_HANDLER_IDENTIFY,
    CLUSTER_HANDLER_LEVEL,
    CLUSTER_HANDLER_ON_OFF,
)
from zhaws.server.zigbee.cluster.general import LevelChangeEvent

if TYPE_CHECKING:
    from zhaws.server.zigbee.cluster import ClusterHandler
    from zhaws.server.zigbee.device import Device
    from zhaws.server.zigbee.endpoint import Endpoint
    from zhaws.server.zigbee.group import Group

STRICT_MATCH = functools.partial(PLATFORM_ENTITIES.strict_match, Platform.LIGHT)
GROUP_MATCH = functools.partial(PLATFORM_ENTITIES.group_match, Platform.LIGHT)

CAPABILITIES_COLOR_LOOP = 0x4
CAPABILITIES_COLOR_XY = 0x08
CAPABILITIES_COLOR_TEMP = 0x10

DEFAULT_TRANSITION = 1

UPDATE_COLORLOOP_ACTION = 0x1
UPDATE_COLORLOOP_DIRECTION = 0x2
UPDATE_COLORLOOP_TIME = 0x4
UPDATE_COLORLOOP_HUE = 0x8

UNSUPPORTED_ATTRIBUTE = 0x86

# Float that represents transition time in seconds to make change.
ATTR_TRANSITION: Final[str] = "transition"

# Lists holding color values
ATTR_RGB_COLOR: Final[str] = "rgb_color"
ATTR_RGBW_COLOR: Final[str] = "rgbw_color"
ATTR_RGBWW_COLOR: Final[str] = "rgbww_color"
ATTR_XY_COLOR: Final[str] = "xy_color"
ATTR_HS_COLOR: Final[str] = "hs_color"
ATTR_COLOR_TEMP: Final[str] = "color_temp"
ATTR_KELVIN: Final[str] = "kelvin"
ATTR_MIN_MIREDS: Final[str] = "min_mireds"
ATTR_MAX_MIREDS: Final[str] = "max_mireds"
ATTR_COLOR_NAME: Final[str] = "color_name"
ATTR_WHITE_VALUE: Final[str] = "white_value"
ATTR_WHITE: Final[str] = "white"

# Brightness of the light, 0..255 or percentage
ATTR_BRIGHTNESS: Final[str] = "brightness"
ATTR_BRIGHTNESS_PCT: Final[str] = "brightness_pct"
ATTR_BRIGHTNESS_STEP: Final[str] = "brightness_step"
ATTR_BRIGHTNESS_STEP_PCT: Final[str] = "brightness_step_pct"

# String representing a profile (built-in ones or external defined).
ATTR_PROFILE: Final[str] = "profile"

# If the light should flash, can be FLASH_SHORT or FLASH_LONG.
ATTR_FLASH: Final[str] = "flash"
FLASH_SHORT: Final[str] = "short"
FLASH_LONG: Final[str] = "long"

# List of possible effects
ATTR_EFFECT_LIST: Final[str] = "effect_list"

# Apply an effect to the light, can be EFFECT_COLORLOOP.
ATTR_EFFECT: Final[str] = "effect"
EFFECT_COLORLOOP: Final[str] = "colorloop"
EFFECT_RANDOM: Final[str] = "random"
EFFECT_WHITE: Final[str] = "white"

ATTR_SUPPORTED_FEATURES: Final[str] = "supported_features"

# Bitfield of features supported by the light entity
SUPPORT_BRIGHTNESS: Final[int] = 1  # Deprecated, replaced by color modes
SUPPORT_COLOR_TEMP: Final[int] = 2  # Deprecated, replaced by color modes
SUPPORT_EFFECT: Final[int] = 4
SUPPORT_FLASH: Final[int] = 8
SUPPORT_COLOR: Final[int] = 16  # Deprecated, replaced by color modes
SUPPORT_TRANSITION: Final[int] = 32
SUPPORT_WHITE_VALUE: Final[int] = 128  # Deprecated, replaced by color modes

EFFECT_BLINK: Final[int] = 0x00
EFFECT_BREATHE: Final[int] = 0x01
EFFECT_OKAY: Final[int] = 0x02
EFFECT_DEFAULT_VARIANT: Final[int] = 0x00

FLASH_EFFECTS: Final[dict[str, int]] = {
    FLASH_SHORT: EFFECT_BLINK,
    FLASH_LONG: EFFECT_BREATHE,
}

SUPPORT_GROUP_LIGHT = (
    SUPPORT_BRIGHTNESS
    | SUPPORT_COLOR_TEMP
    | SUPPORT_EFFECT
    | SUPPORT_FLASH
    | SUPPORT_COLOR
    | SUPPORT_TRANSITION
)


class LightColorMode(enum.IntEnum):
    """ZCL light color mode enum."""

    HS_COLOR = 0x00
    XY_COLOR = 0x01
    COLOR_TEMP = 0x02


class BaseLight(BaseEntity):
    """Operations common to all light entities."""

    PLATFORM = Platform.LIGHT
    _FORCE_ON = False

    def __init__(
        self,
        *args: Any,
        **kwargs: Any,
    ):
        """Initialize the light."""
        super().__init__(*args, **kwargs)
        self._available: bool = False
        self._brightness: Union[int, None] = None
        self._off_brightness: Union[int, None] = None
        self._hs_color: Union[tuple[float, float], None] = None
        self._color_temp: Union[int, None] = None
        self._min_mireds: Union[int, None] = 153
        self._max_mireds: Union[int, None] = 500
        self._effect_list: Union[list[str], None] = None
        self._effect: Union[str, None] = None
        self._supported_features: int = 0
        self._state: bool | None = None
        self._on_off_cluster_handler: ClusterHandler
        self._level_cluster_handler: Optional[ClusterHandler] = None
        self._color_cluster_handler: Optional[ClusterHandler] = None
        self._identify_cluster_handler: Optional[ClusterHandler] = None
        self._default_transition: int | None = None

    def get_state(self) -> dict[str, Any]:
        """Return the state of the light."""
        response = super().get_state()
        response["on"] = self.is_on
        response["brightness"] = self.brightness
        response["hs_color"] = self.hs_color
        response["color_temp"] = self.color_temp
        response["effect"] = self.effect
        response["off_brightness"] = self._off_brightness
        return response

    @property
    def is_on(self) -> bool:
        """Return true if entity is on."""
        if self._state is None:
            return False
        return self._state

    @property
    def brightness(self) -> int | None:
        """Return the brightness of this light."""
        return self._brightness

    @property
    def min_mireds(self) -> int | None:
        """Return the coldest color_temp that this light supports."""
        return self._min_mireds

    @property
    def max_mireds(self) -> int | None:
        """Return the warmest color_temp that this light supports."""
        return self._max_mireds

    def handle_cluster_handler_set_level(self, event: LevelChangeEvent) -> None:
        """Set the brightness of this light between 0..254.
        brightness level 255 is a special value instructing the device to come
        on at `on_level` Zigbee attribute value, regardless of the last set
        level
        """
        value = max(0, min(254, event.level))
        self._brightness = value
        self.maybe_send_state_changed_event()

    @property
    def hs_color(self) -> Optional[tuple[float, float]]:
        """Return the hs color value [int, int]."""
        return self._hs_color

    @property
    def color_temp(self) -> int | None:
        """Return the CT color value in mireds."""
        return self._color_temp

    @property
    def effect_list(self) -> list[str] | None:
        """Return the list of supported effects."""
        return self._effect_list

    @property
    def effect(self) -> str | None:
        """Return the current effect."""
        return self._effect

    @property
    def supported_features(self) -> int:
        """Flag supported features."""
        return self._supported_features

    def to_json(self) -> dict:
        """Return a JSON representation of the select."""
        json = super().to_json()
        json["supported_features"] = self.supported_features
        json["effect_list"] = self.effect_list
        json["min_mireds"] = self.min_mireds
        json["max_mireds"] = self.max_mireds
        return json

    async def async_turn_on(self, transition: int | None = None, **kwargs: Any) -> None:
        """Turn the entity on."""
        duration = (
            transition * 10
            if transition
            else self._default_transition * 10
            if self._default_transition
            else DEFAULT_TRANSITION
        )
        brightness = kwargs.get(ATTR_BRIGHTNESS)
        effect = kwargs.get(ATTR_EFFECT)
        flash = kwargs.get(ATTR_FLASH)

        if brightness is None and self._off_brightness is not None:
            brightness = self._off_brightness

        t_log = {}
        if (
            brightness is not None or transition
        ) and self._supported_features & SUPPORT_BRIGHTNESS:
            if brightness is not None:
                level = min(254, brightness)
            else:
                level = self._brightness or 254
            assert self._level_cluster_handler is not None
            result = await self._level_cluster_handler.move_to_level_with_on_off(
                level, duration
            )
            t_log["move_to_level_with_on_off"] = result
            if not isinstance(result, list) or result[1] is not Status.SUCCESS:
                self.debug("turned on: %s", t_log)
                return
            self._state = bool(level)
            if level:
                self._brightness = level

        if brightness is None or (self._FORCE_ON and brightness):
            # since some lights don't always turn on with move_to_level_with_on_off,
            # we should call the on command on the on_off cluster if brightness is not 0.
            result = await self._on_off_cluster_handler.on()
            t_log["on_off"] = result
            if not isinstance(result, list) or result[1] is not Status.SUCCESS:
                self.debug("turned on: %s", t_log)
                return
            self._state = True
        if ATTR_COLOR_TEMP in kwargs and self.supported_features & SUPPORT_COLOR_TEMP:
            temperature = kwargs[ATTR_COLOR_TEMP]
            assert self._color_cluster_handler is not None
            result = await self._color_cluster_handler.move_to_color_temp(
                temperature, duration
            )
            t_log["move_to_color_temp"] = result
            if not isinstance(result, list) or result[1] is not Status.SUCCESS:
                self.debug("turned on: %s", t_log)
                return
            self._color_temp = temperature
            self._hs_color = None

        if ATTR_HS_COLOR in kwargs and self.supported_features & SUPPORT_COLOR:
            hs_color = kwargs[ATTR_HS_COLOR]
            xy_color = color_util.color_hs_to_xy(*hs_color)
            assert self._color_cluster_handler is not None
            result = await self._color_cluster_handler.move_to_color(
                int(xy_color[0] * 65535), int(xy_color[1] * 65535), duration
            )
            t_log["move_to_color"] = result
            if not isinstance(result, list) or result[1] is not Status.SUCCESS:
                self.debug("turned on: %s", t_log)
                return
            self._hs_color = hs_color
            self._color_temp = None

        if effect == EFFECT_COLORLOOP and self.supported_features & SUPPORT_EFFECT:
            assert self._color_cluster_handler is not None
            result = await self._color_cluster_handler.color_loop_set(
                UPDATE_COLORLOOP_ACTION
                | UPDATE_COLORLOOP_DIRECTION
                | UPDATE_COLORLOOP_TIME,
                0x2,  # start from current hue
                0x1,  # only support up
                transition if transition else 7,  # transition
                0,  # no hue
            )
            t_log["color_loop_set"] = result
            self._effect = EFFECT_COLORLOOP
        elif (
            self._effect == EFFECT_COLORLOOP
            and effect != EFFECT_COLORLOOP
            and self.supported_features & SUPPORT_EFFECT
        ):
            assert self._color_cluster_handler is not None
            result = await self._color_cluster_handler.color_loop_set(
                UPDATE_COLORLOOP_ACTION,
                0x0,
                0x0,
                0x0,
                0x0,  # update action only, action off, no dir, time, hue
            )
            t_log["color_loop_set"] = result
            self._effect = None

        if flash is not None and self._supported_features & SUPPORT_FLASH:
            assert self._identify_cluster_handler is not None
            result = await self._identify_cluster_handler.trigger_effect(
                FLASH_EFFECTS[flash], EFFECT_DEFAULT_VARIANT
            )
            t_log["trigger_effect"] = result

        self._off_brightness = None
        self.debug("turned on: %s", t_log)
        self.maybe_send_state_changed_event()

    async def async_turn_off(
        self, transition: int | None = None, **kwargs: Any
    ) -> None:
        """Turn the entity off."""
        duration = transition
        supports_level = self.supported_features & SUPPORT_BRIGHTNESS

        if duration and supports_level:
            assert self._level_cluster_handler is not None
            result = await self._level_cluster_handler.move_to_level_with_on_off(
                0, duration * 10
            )
        else:
            result = await self._on_off_cluster_handler.off()
        self.debug("turned off: %s", result)
        if not isinstance(result, list) or result[1] is not Status.SUCCESS:
            return
        self._state = False

        if duration and supports_level:
            # store current brightness so that the next turn_on uses it.
            self._off_brightness = self._brightness

        self.maybe_send_state_changed_event()


@STRICT_MATCH(
    cluster_handler_names=CLUSTER_HANDLER_ON_OFF,
    aux_cluster_handlers={CLUSTER_HANDLER_COLOR, CLUSTER_HANDLER_LEVEL},
)
class Light(PlatformEntity, BaseLight):
    """Representation of a light for zhawss."""

    _REFRESH_INTERVAL = (2700, 4500)

    def __init__(
        self,
        unique_id: str,
        cluster_handlers: list[ClusterHandler],
        endpoint: Endpoint,
        device: Device,
    ):
        """Initialize the light."""
        super().__init__(unique_id, cluster_handlers, endpoint, device)
        self._on_off_cluster_handler = self.cluster_handlers[CLUSTER_HANDLER_ON_OFF]
        self._state = bool(self._on_off_cluster_handler.on_off)
        self._level_cluster_handler = self.cluster_handlers.get(CLUSTER_HANDLER_LEVEL)
        self._color_cluster_handler = self.cluster_handlers.get(CLUSTER_HANDLER_COLOR)
        self._identify_cluster_handler = self.cluster_handlers.get(
            CLUSTER_HANDLER_IDENTIFY
        )
        if self._color_cluster_handler:
            self._min_mireds: Union[int, None] = self._color_cluster_handler.min_mireds
            self._max_mireds: Union[int, None] = self._color_cluster_handler.max_mireds
        self._cancel_refresh_handle = None
        effect_list = []

        if self._level_cluster_handler:
            self._supported_features |= SUPPORT_BRIGHTNESS
            self._supported_features |= SUPPORT_TRANSITION
            self._brightness = self._level_cluster_handler.current_level

        if self._color_cluster_handler:
            color_capabilities = self._color_cluster_handler.color_capabilities
            if color_capabilities & CAPABILITIES_COLOR_TEMP:
                self._supported_features |= SUPPORT_COLOR_TEMP
                self._color_temp = self._color_cluster_handler.color_temperature

            if color_capabilities & CAPABILITIES_COLOR_XY:
                self._supported_features |= SUPPORT_COLOR
                curr_x = self._color_cluster_handler.current_x
                curr_y = self._color_cluster_handler.current_y
                if curr_x is not None and curr_y is not None:
                    self._hs_color = color_util.color_xy_to_hs(
                        float(curr_x / 65535), float(curr_y / 65535)
                    )
                else:
                    self._hs_color = (0, 0)

            if color_capabilities & CAPABILITIES_COLOR_LOOP:
                self._supported_features |= SUPPORT_EFFECT
                effect_list.append(EFFECT_COLORLOOP)
                if self._color_cluster_handler.color_loop_active == 1:
                    self._effect = EFFECT_COLORLOOP

        if self._identify_cluster_handler:
            self._supported_features |= SUPPORT_FLASH

        if effect_list:
            self._effect_list = effect_list

        """TODO
        self._default_transition = async_get_zha_config_value(
            zha_device.gateway.config_entry,
            ZHA_OPTIONS,
            CONF_DEFAULT_LIGHT_TRANSITION,
            0,
        )
        """

        self._on_off_cluster_handler.on_event(
            CLUSTER_HANDLER_EVENT, self._handle_event_protocol
        )

        if self._level_cluster_handler:
            self._level_cluster_handler.on_event(
                CLUSTER_HANDLER_EVENT, self._handle_event_protocol
            )

        @periodic(self._REFRESH_INTERVAL)
        async def _refresh() -> None:
            """Call async_get_state at an interval."""
            await self.async_update()
            self.maybe_send_state_changed_event()

        self._cancel_refresh_handle = asyncio.create_task(_refresh())

    def handle_cluster_handler_attribute_updated(
        self, event: ClusterAttributeUpdatedEvent
    ) -> None:
        """Set the state."""
        self._state = bool(event.value)
        if event.value:
            self._off_brightness = None
        self.maybe_send_state_changed_event()

    async def async_update(self) -> None:
        """Attempt to retrieve the state from the light."""
        if not self.available:
            return
        self.debug("polling current state")
        if self._on_off_cluster_handler:
            state = await self._on_off_cluster_handler.get_attribute_value(
                "on_off", from_cache=False
            )
            if state is not None:
                self._state = state
        if self._level_cluster_handler:
            level = await self._level_cluster_handler.get_attribute_value(
                "current_level", from_cache=False
            )
            if level is not None:
                self._brightness = level
        if self._color_cluster_handler:
            attributes = [
                "color_mode",
                "color_temperature",
                "current_x",
                "current_y",
                "color_loop_active",
            ]

            results = await self._color_cluster_handler.get_attributes(
                attributes, from_cache=False
            )

            if (color_mode := results.get("color_mode")) is not None:
                if color_mode == LightColorMode.COLOR_TEMP:
                    color_temp = results.get("color_temperature")
                    if color_temp is not None and color_mode:
                        self._color_temp = color_temp
                        self._hs_color = None
                else:
                    color_x = results.get("current_x")
                    color_y = results.get("current_y")
                    if color_x is not None and color_y is not None:
                        self._hs_color = color_util.color_xy_to_hs(
                            float(color_x / 65535), float(color_y / 65535)
                        )
                        self._color_temp = None

            color_loop_active = results.get("color_loop_active")
            if color_loop_active is not None:
                if color_loop_active == 1:
                    self._effect = EFFECT_COLORLOOP
                else:
                    self._effect = None


@STRICT_MATCH(
    cluster_handler_names=CLUSTER_HANDLER_ON_OFF,
    aux_cluster_handlers={CLUSTER_HANDLER_COLOR, CLUSTER_HANDLER_LEVEL},
    manufacturers={"Philips", "Signify Netherlands B.V."},
)
class HueLight(Light):
    """Representation of a HUE light which does not report attributes."""

    _REFRESH_INTERVAL = (180, 300)


@STRICT_MATCH(
    cluster_handler_names=CLUSTER_HANDLER_ON_OFF,
    aux_cluster_handlers={CLUSTER_HANDLER_COLOR, CLUSTER_HANDLER_LEVEL},
    manufacturers={"Jasco", "Quotra-Vision"},
)
class ForceOnLight(Light):
    """Representation of a light which does not respect move_to_level_with_on_off."""

    _FORCE_ON = True


@GROUP_MATCH()
class LightGroup(GroupEntity, BaseLight):
    """Representation of a light group."""

    def __init__(self, group: Group):
        """Initialize a light group."""
        super().__init__(group)
        self._on_off_cluster_handler: ClusterHandler = group.zigpy_group.endpoint[
            OnOff.cluster_id
        ]
        self._level_cluster_handler: Optional[
            ClusterHandler
        ] = group.zigpy_group.endpoint[LevelControl.cluster_id]
        self._color_cluster_handler: Optional[
            ClusterHandler
        ] = group.zigpy_group.endpoint[Color.cluster_id]
        self._identify_cluster_handler: Optional[
            ClusterHandler
        ] = group.zigpy_group.endpoint[Identify.cluster_id]

    def update(self, _: Any = None) -> None:
        # Query all members and determine the light group state.
        self.debug("Updating light group entity state")
        platform_entities = self._group.get_platform_entities(self.PLATFORM)
        all_entities = [entity.to_json() for entity in platform_entities]
        all_states = [entity["state"] for entity in all_entities]
        self.debug(
            "All platform entity states for group entity members: %s", all_states
        )
        on_states = [state for state in all_states if state["on"]]

        self._state = len(on_states) > 0

        self._available = any(entity.available for entity in platform_entities)

        self._brightness = helpers.reduce_attribute(on_states, ATTR_BRIGHTNESS)

        self._hs_color = helpers.reduce_attribute(
            on_states, ATTR_HS_COLOR, reduce=helpers.mean_tuple
        )

        self._color_temp = helpers.reduce_attribute(on_states, ATTR_COLOR_TEMP)
        self._min_mireds = helpers.reduce_attribute(
            all_entities, ATTR_MIN_MIREDS, default=153, reduce=min
        )
        self._max_mireds = helpers.reduce_attribute(
            all_entities, ATTR_MAX_MIREDS, default=500, reduce=max
        )

        self._effect_list = None
        all_effect_lists = list(
            helpers.find_state_attributes(all_entities, ATTR_EFFECT_LIST)
        )
        if all_effect_lists:
            # Merge all effects from all effect_lists with a union merge.
            self._effect_list = list(set().union(*all_effect_lists))

        self._effect = None
        all_effects = list(helpers.find_state_attributes(all_states, ATTR_EFFECT))
        if all_effects:
            # Report the most common effect.
            effects_count = Counter(itertools.chain(all_effects))
            self._effect = effects_count.most_common(1)[0][0]

        self._supported_features = 0
        for support in helpers.find_state_attributes(
            all_entities, ATTR_SUPPORTED_FEATURES
        ):
            # Merge supported features by emulating support for every feature
            # we find.
            self._supported_features |= support
        # Bitwise-and the supported features with the GroupedLight's features
        # so that we don't break in the future when a new feature is added.
        self._supported_features &= SUPPORT_GROUP_LIGHT

        self.maybe_send_state_changed_event()



"""Models that represent messages in zhawss."""
from typing import Annotated, Union

from pydantic.fields import Field

from zhaws.client.model.commands import CommandResponses
from zhaws.client.model.events import Events
from zhaws.model import BaseModel


class Message(BaseModel):
    """Response model."""

    __root__: Annotated[
        Union[CommandResponses, Events],
        Field(discriminator="message_type"),  # noqa: F821
    ]



"""Controller for zha web socket server."""
from __future__ import annotations

import asyncio
import logging
import time
from typing import TYPE_CHECKING, Any, Optional, Union

from bellows.zigbee.application import ControllerApplication
from serial.serialutil import SerialException
from zhaquirks import setup as setup_quirks
from zigpy.endpoint import Endpoint
from zigpy.group import Group as ZigpyGroup
from zigpy.types.named import EUI64
from zigpy.typing import DeviceType as ZigpyDeviceType

from zhaws.backports.enum import StrEnum
from zhaws.server.const import (
    CONF_ENABLE_QUIRKS,
    CONF_RADIO_TYPE,
    DEVICE,
    EVENT,
    EVENT_TYPE,
    IEEE,
    MESSAGE_TYPE,
    NWK,
    PAIRING_STATUS,
    ControllerEvents,
    EventTypes,
    MessageTypes,
)
from zhaws.server.platforms import discovery
from zhaws.server.zigbee.group import Group, GroupMemberReference

if TYPE_CHECKING:
    from zhaws.server.websocket.server import Server

from zhaws.server.zigbee.device import Device, DeviceStatus
from zhaws.server.zigbee.radio import RadioType

_LOGGER = logging.getLogger(__name__)


class DevicePairingStatus(StrEnum):
    """Status of a device."""

    PAIRED = "paired"
    INTERVIEW_COMPLETE = "interview_complete"
    CONFIGURED = "configured"
    INITIALIZED = "initialized"


class Controller:
    """Controller for the Zigbee application."""

    def __init__(self, server: Server):
        """Initialize the controller."""
        self._application_controller: ControllerApplication | None = None
        self._server: Server = server
        self.radio_description: Optional[str] = None
        self._devices: dict[EUI64, Device] = {}
        self._groups: dict[int, Group] = {}

    @property
    def is_running(self) -> bool:
        """Return true if the controller is running."""
        return self._application_controller is not None

    @property
    def server(self) -> Server:
        """Return the server."""
        return self._server

    @property
    def application_controller(self) -> ControllerApplication:
        """Return the Zigpy ControllerApplication"""
        return self._application_controller

    @property
    def coordinator_device(self) -> Device:
        """Get the coordinator device."""
        return self._devices[self._application_controller.ieee]

    @property
    def devices(self) -> dict[EUI64, Device]:
        """Get devices."""
        return self._devices

    @property
    def groups(self) -> dict[int, Group]:
        """Get groups."""
        return self._groups

    async def start_network(self, configuration: dict) -> None:
        """Start the Zigbee network."""
        if configuration.get(CONF_ENABLE_QUIRKS):
            setup_quirks(configuration)
        radio_type = configuration[CONF_RADIO_TYPE]
        app_controller_cls = RadioType[radio_type].controller
        self.radio_description = RadioType[radio_type].description
        controller_config = app_controller_cls.SCHEMA(configuration)  # type: ignore
        try:
            self._application_controller = await app_controller_cls.new(  # type: ignore
                controller_config, auto_form=True, start_radio=True
            )
        except (asyncio.TimeoutError, SerialException, OSError) as exc:
            _LOGGER.error(
                "Couldn't start %s coordinator",
                self.radio_description,
                exc_info=exc,
            )
            raise
        self.load_devices()
        self.load_groups()
        self.application_controller.add_listener(self)
        self.application_controller.groups.add_listener(self)

    def load_devices(self) -> None:
        """Load devices."""
        self._devices = {
            zigpy_device.ieee: Device(zigpy_device, self)
            for zigpy_device in self._application_controller.devices.values()
        }
        self.create_platform_entities()

    def load_groups(self) -> None:
        """Load groups."""
        for group_id, group in self.application_controller.groups.items():
            _LOGGER.info("Loading group with id: 0x%04x", group_id)
            group = Group(group, self._server)
            self._groups[group_id] = group
            # we can do this here because the entities are in the entity registry tied to the devices
            discovery.GROUP_PROBE.discover_group_entities(group)

    def create_platform_entities(self) -> None:
        """Create platform entities."""

        for platform in discovery.PLATFORMS:
            for platform_entity_class, args in self.server.data[platform]:
                platform_entity = platform_entity_class.create_platform_entity(*args)
                if platform_entity:
                    _LOGGER.debug("Platform entity data: %s", platform_entity.to_json())
            self.server.data[platform].clear()

    async def stop_network(self) -> None:
        """Stop the Zigbee network."""
        if self._application_controller is None:
            return

        await self._application_controller.pre_shutdown()
        self._application_controller = None

    def get_device(self, ieee: Union[EUI64, str]) -> Device:
        """Get a device by ieee address."""
        if isinstance(ieee, str):
            ieee = EUI64.convert(ieee)
        device = self._devices.get(ieee)
        if not device:
            raise ValueError(f"Device {str(ieee)} not found")
        return device

    def get_group(self, group_id: int) -> Group:
        """Get a group by group id."""
        group = self._groups.get(group_id)
        if not group:
            raise ValueError(f"Group {str(group_id)} not found")
        return group

    def device_joined(self, device: ZigpyDeviceType) -> None:
        """Handle device joined.

        At this point, no information about the device is known other than its
        address
        """
        _LOGGER.info("Device %s - %s joined", device.ieee, f"0x{device.nwk:04x}")
        self.server.client_manager.broadcast(
            {
                MESSAGE_TYPE: MessageTypes.EVENT,
                EVENT_TYPE: EventTypes.CONTROLLER_EVENT,
                EVENT: ControllerEvents.DEVICE_JOINED,
                IEEE: str(device.ieee),
                NWK: f"0x{device.nwk:04x}",
                PAIRING_STATUS: DevicePairingStatus.PAIRED,
            }
        )

    def raw_device_initialized(self, device: ZigpyDeviceType) -> None:
        """Handle a device initialization without quirks loaded."""
        _LOGGER.info(
            "Device %s - %s raw device initialized", device.ieee, f"0x{device.nwk:04x}"
        )

        self.server.client_manager.broadcast(
            {
                MESSAGE_TYPE: MessageTypes.EVENT,
                EVENT_TYPE: EventTypes.CONTROLLER_EVENT,
                EVENT: ControllerEvents.RAW_DEVICE_INITIALIZED,
                IEEE: str(device.ieee),
                NWK: f"0x{device.nwk:04x}",
                PAIRING_STATUS: DevicePairingStatus.INTERVIEW_COMPLETE,
                "model": device.model if device.model else "unknown_model",
                "manufacturer": device.manufacturer
                if device.manufacturer
                else "unknown_manufacturer",
                "signature": device.get_signature(),
            }
        )

    def device_initialized(self, device: ZigpyDeviceType) -> None:
        """Handle device joined and basic information discovered."""
        _LOGGER.info("Device %s - %s initialized", device.ieee, f"0x{device.nwk:04x}")
        asyncio.create_task(self.async_device_initialized(device))

    def device_left(self, device: ZigpyDeviceType) -> None:
        """Handle device leaving the network."""
        _LOGGER.info("Device %s - %s left", device.ieee, f"0x{device.nwk:04x}")
        self.server.client_manager.broadcast(
            {
                MESSAGE_TYPE: MessageTypes.EVENT,
                EVENT_TYPE: EventTypes.CONTROLLER_EVENT,
                EVENT: ControllerEvents.DEVICE_LEFT,
                IEEE: str(device.ieee),
                NWK: f"0x{device.nwk:04x}",
            }
        )

    def device_removed(self, device: ZigpyDeviceType) -> None:
        """Handle device being removed from the network."""
        device = self._devices.pop(device.ieee, None)
        if device is not None:
            message: dict[str, Any] = {DEVICE: device.zha_device_info}
            message[MESSAGE_TYPE] = MessageTypes.EVENT
            message[EVENT_TYPE] = EventTypes.CONTROLLER_EVENT
            message[EVENT] = ControllerEvents.DEVICE_REMOVED
            self.server.client_manager.broadcast(message)

    def group_member_removed(self, zigpy_group: ZigpyGroup, endpoint: Endpoint) -> None:
        """Handle zigpy group member removed event."""
        # need to handle endpoint correctly on groups
        group = self.get_or_create_group(zigpy_group)
        group.info("group_member_removed - endpoint: %s", endpoint)
        self._emit_group_event(group, ControllerEvents.GROUP_MEMBER_REMOVED)

    def group_member_added(self, zigpy_group: ZigpyGroup, endpoint: Endpoint) -> None:
        """Handle zigpy group member added event."""
        # need to handle endpoint correctly on groups
        group = self.get_or_create_group(zigpy_group)
        group.info("group_member_added - endpoint: %s", endpoint)
        self._emit_group_event(group, ControllerEvents.GROUP_MEMBER_ADDED)
        if len(group.members) > 1:
            # we need to do this because there wasn't already a group entity to remove and re-add
            discovery.GROUP_PROBE.discover_group_entities(group)

    def group_added(self, zigpy_group: ZigpyGroup) -> None:
        """Handle zigpy group added event."""
        group = self.get_or_create_group(zigpy_group)
        group.info("group_added")
        self._emit_group_event(group, ControllerEvents.GROUP_ADDED)

    def group_removed(self, zigpy_group: ZigpyGroup) -> None:
        """Handle zigpy group removed event."""
        # TODO handle entity change unsubs and sub for new events
        group = self._groups.pop(zigpy_group.group_id, None)
        if group is not None:
            group.info("group_removed")
            self._emit_group_event(group, ControllerEvents.GROUP_REMOVED)

    async def async_device_initialized(self, device: ZigpyDeviceType) -> None:
        """Handle device joined and basic information discovered (async)."""
        zha_device = self.get_or_create_device(device)
        # This is an active device so set a last seen if it is none
        if zha_device.last_seen is None:
            zha_device.async_update_last_seen(time.time())
        _LOGGER.debug(
            "device - %s:%s entering async_device_initialized - is_new_join: %s",
            f"0x{device.nwk:04x}",
            device.ieee,
            zha_device.status is not DeviceStatus.INITIALIZED,
        )

        if zha_device.status is DeviceStatus.INITIALIZED:
            # ZHA already has an initialized device so either the device was assigned a
            # new nwk or device was physically reset and added again without being removed
            _LOGGER.debug(
                "device - %s:%s has been reset and re-added or its nwk address changed",
                f"0x{device.nwk:04x}",
                device.ieee,
            )
            await self._async_device_rejoined(zha_device)
        else:
            _LOGGER.debug(
                "device - %s:%s has joined the ZHA zigbee network",
                f"0x{device.nwk:04x}",
                device.ieee,
            )
            await self._async_device_joined(zha_device)

        message: dict[str, Any] = {DEVICE: zha_device.zha_device_info}
        message[PAIRING_STATUS] = DevicePairingStatus.INITIALIZED
        message[MESSAGE_TYPE] = MessageTypes.EVENT
        message[EVENT_TYPE] = EventTypes.CONTROLLER_EVENT
        message[EVENT] = ControllerEvents.DEVICE_FULLY_INITIALIZED
        self.server.client_manager.broadcast(message)

    def get_or_create_device(self, zigpy_device: ZigpyDeviceType) -> Device:
        """Get or create a device."""
        if (device := self._devices.get(zigpy_device.ieee)) is None:
            device = Device(zigpy_device, self)
            self._devices[zigpy_device.ieee] = device
        return device

    def get_or_create_group(self, zigpy_group: ZigpyGroup) -> Group:
        """Get or create a group."""
        group = self._groups.get(zigpy_group.group_id)
        if group is None:
            group = Group(zigpy_group, self.server)
            self._groups[zigpy_group.group_id] = group
        return group

    async def _async_device_joined(self, device: Device) -> None:
        device.available = True
        message: dict[str, Any] = {DEVICE: device.device_info}
        await device.async_configure()
        message[PAIRING_STATUS] = DevicePairingStatus.CONFIGURED
        message[MESSAGE_TYPE] = MessageTypes.EVENT
        message[EVENT_TYPE] = EventTypes.CONTROLLER_EVENT
        message[EVENT] = ControllerEvents.DEVICE_CONFIGURED
        self.server.client_manager.broadcast(message)
        await device.async_initialize(from_cache=False)
        self.create_platform_entities()

    async def _async_device_rejoined(self, device: Device) -> None:
        _LOGGER.debug(
            "skipping discovery for previously discovered device - %s:%s",
            f"0x{device.nwk:04x}",
            device.ieee,
        )
        # we don't have to do this on a nwk swap but we don't have a way to tell currently
        await device.async_configure()
        message: dict[str, Any] = {DEVICE: device.device_info}
        message[PAIRING_STATUS] = DevicePairingStatus.CONFIGURED
        message[MESSAGE_TYPE] = MessageTypes.EVENT
        message[EVENT_TYPE] = EventTypes.CONTROLLER_EVENT
        message[EVENT] = ControllerEvents.DEVICE_CONFIGURED
        self.server.client_manager.broadcast(message)
        # force async_initialize() to fire so don't explicitly call it
        device.available = False
        device.update_available(True)

    def get_group_by_name(self, group_name: str) -> Group | None:
        """Get ZHA group by name."""
        for group in self._groups.values():
            if group.name == group_name:
                return group
        return None

    async def async_create_zigpy_group(
        self,
        name: str,
        members: list[GroupMemberReference],
        group_id: int | None = None,
    ) -> Group:
        """Create a new Zigpy Zigbee group."""
        # we start with two to fill any gaps from a user removing existing groups

        if group_id is None:
            group_id = 2
            while group_id in self._groups:
                group_id += 1

        # guard against group already existing
        if self.get_group_by_name(name) is None:
            self.application_controller.groups.add_group(group_id, name)
            if members is not None:
                tasks = []
                for member in members:
                    _LOGGER.debug(
                        "Adding member with IEEE: %s and endpoint ID: %s to group: %s:0x%04x",
                        member.ieee,
                        member.endpoint_id,
                        name,
                        group_id,
                    )
                    tasks.append(
                        self.devices[member.ieee].async_add_endpoint_to_group(
                            member.endpoint_id, group_id
                        )
                    )
                await asyncio.gather(*tasks)
        return self._groups[group_id]

    async def async_remove_zigpy_group(self, group_id: int) -> None:
        """Remove a Zigbee group from Zigpy."""
        if not (group := self._groups.get(group_id)):
            _LOGGER.debug("Group: 0x%04x could not be found", group_id)
            return
        if group.members:
            tasks = []
            for member in group.members:
                tasks.append(member.async_remove_from_group())
            if tasks:
                await asyncio.gather(*tasks)
        self.application_controller.groups.pop(group_id)

    def _emit_group_event(self, group: Group, event: str) -> None:
        """Emit group event."""
        message: dict[str, Any] = {"group": group.to_json()}
        message[MESSAGE_TYPE] = MessageTypes.EVENT
        message[EVENT_TYPE] = EventTypes.CONTROLLER_EVENT
        message[EVENT] = event
        self.server.client_manager.broadcast(message)



"""Zigbee radio utilities."""

import enum
from typing import Callable, Final

import bellows.zigbee.application
from zigpy.config import CONF_DEVICE_PATH  # noqa: F401 # pylint: disable=unused-import
import zigpy_deconz.zigbee.application
import zigpy_xbee.zigbee.application
import zigpy_zigate.zigbee.application
import zigpy_znp.zigbee.application

BAUD_RATES: Final[list[int]] = [
    2400,
    4800,
    9600,
    14400,
    19200,
    38400,
    57600,
    115200,
    128000,
    256000,
]


class RadioType(enum.Enum):
    """Possible options for radio type."""

    znp = (
        "ZNP = Texas Instruments Z-Stack ZNP protocol: CC253x, CC26x2, CC13x2",
        zigpy_znp.zigbee.application.ControllerApplication,
    )
    ezsp = (
        "EZSP = Silicon Labs EmberZNet protocol: Elelabs, HUSBZB-1, Telegesis",
        bellows.zigbee.application.ControllerApplication,
    )
    deconz = (
        "deCONZ = dresden elektronik deCONZ protocol: ConBee I/II, RaspBee I/II",
        zigpy_deconz.zigbee.application.ControllerApplication,
    )
    zigate = (
        "ZiGate = ZiGate Zigbee radios: PiZiGate, ZiGate USB-TTL, ZiGate WiFi",
        zigpy_zigate.zigbee.application.ControllerApplication,
    )
    xbee = (
        "XBee = Digi XBee Zigbee radios: Digi XBee Series 2, 2C, 3",
        zigpy_xbee.zigbee.application.ControllerApplication,
    )

    @classmethod
    def list(cls) -> list[str]:
        """Return a list of descriptions."""
        return [e.description for e in RadioType]

    @classmethod
    def get_by_description(cls, description: str) -> str:
        """Get radio by description."""
        for radio in cls:
            if radio.description == description:
                return radio.name
        raise ValueError

    def __init__(self, description: str, controller_cls: Callable) -> None:
        """Init instance."""
        self._desc = description
        self._ctrl_cls = controller_cls

    @property
    def controller(self) -> Callable:
        """Return controller class."""
        return self._ctrl_cls

    @property
    def description(self) -> str:
        """Return radio type description."""
        return self._desc



"""Lighting cluster handlers module for zhawss."""
from __future__ import annotations

from contextlib import suppress

from zigpy.zcl.clusters import lighting

from zhaws.server.zigbee import registries
from zhaws.server.zigbee.cluster import ClientClusterHandler, ClusterHandler
from zhaws.server.zigbee.cluster.const import REPORT_CONFIG_DEFAULT


@registries.CLUSTER_HANDLER_REGISTRY.register(lighting.Ballast.cluster_id)
class Ballast(ClusterHandler):
    """Ballast cluster handler."""


@registries.CLIENT_CLUSTER_HANDLER_REGISTRY.register(lighting.Color.cluster_id)
class ColorClientClusterHandler(ClientClusterHandler):
    """Color client cluster handler."""


@registries.BINDABLE_CLUSTERS.register(lighting.Color.cluster_id)
@registries.CLUSTER_HANDLER_REGISTRY.register(lighting.Color.cluster_id)
class ColorClusterHandler(ClusterHandler):
    """Color cluster handler."""

    CAPABILITIES_COLOR_XY = 0x08
    CAPABILITIES_COLOR_TEMP = 0x10
    UNSUPPORTED_ATTRIBUTE = 0x86
    REPORT_CONFIG = [
        {"attr": "current_x", "config": REPORT_CONFIG_DEFAULT},
        {"attr": "current_y", "config": REPORT_CONFIG_DEFAULT},
        {"attr": "color_temperature", "config": REPORT_CONFIG_DEFAULT},
    ]
    MAX_MIREDS: int = 500
    MIN_MIREDS: int = 153
    ZCL_INIT_ATTRS = {
        "color_temp_physical_min": True,
        "color_temp_physical_max": True,
        "color_capabilities": True,
        "color_loop_active": False,
    }

    @property
    def color_capabilities(self) -> int:
        """Return color capabilities of the light."""
        with suppress(KeyError):
            return self.cluster["color_capabilities"]
        if self.cluster.get("color_temperature") is not None:
            return self.CAPABILITIES_COLOR_XY | self.CAPABILITIES_COLOR_TEMP
        return self.CAPABILITIES_COLOR_XY

    @property
    def color_loop_active(self) -> int | None:
        """Return cached value of the color_loop_active attribute."""
        return self.cluster.get("color_loop_active")

    @property
    def color_temperature(self) -> int | None:
        """Return cached value of color temperature."""
        return self.cluster.get("color_temperature")

    @property
    def current_x(self) -> int | None:
        """Return cached value of the current_x attribute."""
        return self.cluster.get("current_x")

    @property
    def current_y(self) -> int | None:
        """Return cached value of the current_y attribute."""
        return self.cluster.get("current_y")

    @property
    def min_mireds(self) -> int:
        """Return the coldest color_temp that this cluster handler supports."""
        return self.cluster.get("color_temp_physical_min", self.MIN_MIREDS)

    @property
    def max_mireds(self) -> int:
        """Return the warmest color_temp that this cluster handler supports."""
        return self.cluster.get("color_temp_physical_max", self.MAX_MIREDS)
